import itertools
import pandas as pd
import numpy as np
import copy
from tqdm import tqdm
from sklearn.model_selection import KFold

from cwgp.cwgp import CWGP


def grid_search(
        estimator,
        x,
        y,
        params={},
        **kwargs):
    """
    Exhausts all given combinations of cwgp

    Parameters:
    ----------
    estimator: func
            a user defined function that uses x, the independent variable of the data, and the cwgp transformation of y,
            the dependent variable of the data.
    x : np.array
        the independent variable of the data
    y : np.array
        the dependent variable of the data, the target of the transformation
    params : dict
        the range of grid search
        >>> {"c":2,"n":[1,2,3],"transformations":["sa","sal"]}
    Returns
    -------
    cwpg : dict
        A dictionary containing all of the results
    """
    c = params.get("c", 2)
    n = params.get("n", [2])
    transformations = params.get("transformations", ["sa", "sal"])
    cwgp_params = [transformations, n]
    params_product = list(itertools.product(*cwgp_params))
    params_combination = list(itertools.product(params_product, repeat=c))
    print(params_product)

    n_splits = kwargs.get("n_splits", 10)
    cv = kwargs.get("cv", False)
    shuffle = kwargs.get("shuffle", False)
    random_state = kwargs.get("random_state", 42)
    reverse_model_order = kwargs.get("reverse_model_order", False)

    cwgp = {}
    for index, param in enumerate(tqdm(params_combination)):
        cwgp[index] = {}

        t_data = copy.deepcopy(y)
        if cv:
            kf = KFold(
                n_splits=n_splits,
                random_state=random_state,
                shuffle=shuffle)
            for split_index, (train, val) in enumerate(kf.split(t_data)):
                x_train, x_val = x[train], x[val]
                y_train, y_val = t_data[train], t_data[val]
                model_holder = fit_transform(
                    param, y_train, reverse_model_order=reverse_model_order)
                cwgp[index][split_index] = estimator(
                    x_train=x_train,
                    y_train=y_train,
                    x_val=x_val,
                    y_val=y_val,
                    x=x,
                    y=t_data,
                    train=train,
                    val=val,
                    hyperparams=param,
                    model_holder=model_holder,
                    **kwargs)
        else:
            model_holder = fit_transform(
                param, t_data, reverse_model_order=reverse_model_order)
            cwgp[index]["result"] = estimator(
                x_train=x,
                y_train=t_data,
                model_holder=model_holder,
                hyperparams=param,
                **kwargs)
        cwgp[index]["cwgp_combination"] = param
    return cwgp


def fit_transform(param, y, reverse_model_order=False):
    model_holder = []
    for t, d in param:
        cwgp_model = CWGP(t, n=d)
        cwgp_model.fit(y)
        y, y_d = cwgp_model.phi.comp_phi(
            cwgp_model.phi.res.x, y)
        model_holder.append(cwgp_model)
    if reverse_model_order:
        # For conveniences in inverse computation
        model_holder = model_holder[::-1]
    return model_holder



import numpy as np


def distance(x, y, metric="Euclidean"):
    if metric == "Euclidean":
        return np.abs(np.subtract.outer(x.reshape(-1), y.reshape(-1)))


def OU(x, y, par=[1]):
    return np.exp(-par[0] *
                  distance(x, y))


def RBF(x, y, par=[1]):
    return np.exp(-
                  np.power(distance(x / (par[0]), y / (par[0])), 2))


def Matern32(x, y, par=[1, 1]):
    return par[0] * (1 + np.sqrt(3) * distance(x, y) / par[1]) * \
        np.exp(-np.sqrt(3) * distance(x, y) / par[1])



from cwgp.phi import Phi


class CWGP():
    def __init__(self, fn, *args, **kwargs):
        self.phi = Phi(fn, *args, **kwargs)

    def fit(self, y, **kwargs):
        return self.phi.minimize_lf(y, **kwargs)

    def normality_test(self):
        pass



import autograd.numpy as np


def sal(par, y):
    return par[0] + par[1] * np.sinh(par[2] * np.arcsinh(y) + par[3])


def inv_sal(par, x):
    return np.sinh((np.arcsinh((x - par[0]) / par[1]) - par[3]) / par[2])


def asinh(par, y):
    return par[0] + par[1] * np.arcsinh((y - par[2]) / par[3])


def inv_asinh(par, x):
    return ((np.sinh(x - par[0]) / par[1]) * par[3]) + par[2]


def box_cox(par, y):
    return (np.sign(y) * np.abs(y)**par[0] - 1) / par[0]


def inv_box_cox(par, x):
    return np.sign(par[0] * x + 1) * (np.abs(par[0] * x + 1)**(1 / par[0]))


def sa(par, y):
    return np.sinh(par[0] * np.arcsinh(y) + par[1])


def inv_sa(par, x):
    return np.sinh((np.arcsinh(x) - par[1]) / par[0])



from autograd import elementwise_grad, grad
import autograd.numpy as np
import copy
from scipy.optimize import minimize

from cwgp.kernel import RBF, OU, Matern32
from cwgp.transformations import sal, sa, asinh, box_cox, inv_sal, inv_sa, inv_asinh, inv_box_cox


class Phi():
    FN_BANK = {
        "sal": {"fn": sal, "inv_fn": inv_sal, "par_len": 4},
        "sa": {"fn": sa, "inv_fn": inv_sa, "par_len": 2},
        "asinh": {"fn": asinh, "inv_fn": inv_asinh, "par_len": 4},
        "box_cox": {"fn": box_cox, "inv_fn": inv_box_cox, "par_len": 1},
    }

    KERNEL_BANK = {
        "OU": {"kern": OU, "params": 1, "init_scale": 1},
        "RBF": {"kern": RBF, "params": 1, "init_scale": 10},
        "Matern32": {"kern": Matern32, "params": 2, "init_scale": 10},
    }

    def __init__(
            self,
            fn,
            kernel="OU",
            kernel_params_estimate=False,
            par_len=None,
            n=1):
        self.fn = self.FN_BANK[fn]["fn"]  # a differentiable function
        self.inv_fn = self.FN_BANK[fn]["inv_fn"]
        self.d_fn = elementwise_grad(self.fn, 1)  # take derivative
        self.kernel = self.KERNEL_BANK[kernel]["kern"]
        self.kernel_params = self.KERNEL_BANK[kernel]["params"] if kernel_params_estimate else 0
        self.par_len = self.FN_BANK[fn]["par_len"]
        self.init_scale = self.KERNEL_BANK[kernel]["init_scale"]
        self.n = n

    def comp_phi(self, par, y):
        assert len(par) >= self.par_len * self.n, "Not enough parameters"
        comp = copy.deepcopy(y)
        d_comp = 1
        for i in range(0, self.n):
            d_comp *= self.d_fn(par[self.par_len * i:], comp)
            comp = self.fn(par[self.par_len * i:], comp)
        return comp, d_comp

    def inv_comp_phi(self, par, x):
        assert len(par) >= self.par_len * self.n, "Not enough parameters"
        if self.kernel_params:
            par = par[:-self.kernel_params]
        inv_comp = copy.deepcopy(x)
        for i in range(0, self.n):
            if i == 0:
                inv_comp = self.inv_fn(par[-self.par_len:], inv_comp)
            else:
                inv_comp = self.inv_fn(
                    par[-self.par_len * (i + 1):-self.par_len * i], inv_comp)
        return inv_comp

    def likelihood(self, par, y):
        phi_y, chain_d_sal = self.comp_phi(par, y)
        t_phi_y = np.transpose(phi_y)
        if self.kernel_params:
            cov_xx = self.kernel(phi_y, t_phi_y, par[-self.kernel_params:])
        else:
            cov_xx = self.kernel(phi_y, t_phi_y)
        gaussian_params = 0.5 * (t_phi_y) @ np.linalg.inv(cov_xx) @ phi_y
        return np.ravel(0.5 * np.log(np.linalg.det(cov_xx)) +
                        gaussian_params - sum(np.log(chain_d_sal)))

    def reml(self, method):
        pass

    def minimize_lf(self, y, method='l-bfgs-b', loop=True):
        # http://stackoverflow.com/questions/19843752/structure-of-inputs-to-scipy-minimize-function
        res = minimize
        res.success = False
        if loop:
            while res.success == False:
                try:
                    res = minimize(
                        self.likelihood,
                        self.init_scale * np.random.rand(
                            self.par_len *
                            self.n +
                            self.kernel_params),
                        args=(
                            y,
                        ),
                        method=method)
                except Exception:
                    pass
        self.res = res
        return res



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="CWGP",
    version="1.4.5",
    author="Andy Chen",
    author_email="andy97_861022_chen@hotmail.com",
    description="CWGP",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/andy971022/CWGP",
    packages=setuptools.find_packages(),
    classifiers=[
                "Programming Language :: Python :: 3",
                "License :: OSI Approved :: MIT License",
                "Operating System :: OS Independent",
    ],
    install_requires=[
        'autograd',
        'scipy>=1.5.4',
        'tqdm>=4.56.1',
        'numpy>=1.18.0'
    ],
    python_requires='>=3.6',
)
